︠4fdb2723-ab40-433d-b632-4aa903f263a0i︠
%md
# Math 152: Intro to Mathematical Software
### 2017-02-08
### Kiran Kedlaya; University of California, San Diego
### adapted from lectures by William Stein, University of Washington


## **Lecture 13: Timing and Speed**
︡e62a0d2b-7048-470e-b0c1-c157ea4c0265︡{"done":true,"md":"# Math 152: Intro to Mathematical Software\n### 2017-02-08\n### Kiran Kedlaya; University of California, San Diego\n### adapted from lectures by William Stein, University of Washington\n\n\n## **Lecture 13: Timing and Speed**"}
︠67ef4bf4-e360-4ef2-aeff-8d1909536bfbi︠
%md
## Timing your code

Here's some code.  It is intentionally naive in various ways...
︡ce6fbcae-c546-48d9-be6c-3fcbdc96866c︡{"done":true,"md":"## Timing your code\n\nHere's some code.  It is intentionally naive in various ways..."}
︠512e0ece-978e-4018-ae69-981d33d8c474s︠
def sin_sum(n):
    """
    Return the "sum"  sin(1) + sin(2) + ... + sin(n-1)
    """
    return sum(float(sin(m)) for m in xrange(1,n))
︡605b7a2b-fc0f-4cee-b7db-f26d07d123d4︡{"done":true}︡
︠645441f2-d2d5-494f-9c9d-0e9c46a8a1ee︠
︡f67f8b22-59a8-4b43-a20c-ad0aa7b0da4c︡
︠623ff2e1-0e54-4869-89d0-f2f343d0df93s︠
sin_sum(10000)
︡41167fc9-7561-4b3b-8f8f-e9866f7d852a︡{"stdout":"1.9395054106806988"}︡{"stdout":"\n"}︡{"done":true}︡
︠bbd16d90-e351-4a09-8fee-7133e4f3ff62i︠
%md
Explain this code line by line:

- define function of one input
- Docstring
- sin(m) is the sage symbolic sin function
- `(sin(m) for m in xrange(1,n))` is like a list comprehension but it is "lazy" so it doesn't waste memory -- you can iterate over it though.
- `sum(...)` iterates over its input summing it up
- `float(...)` converts its input to a float


Throw out a mathematical question: Is there a simple formula for $\sum_{0< m\leq n} \sin(m)$ similar to the formula
$\sum_{0< m<n} m = \frac{n(n-1)}{2}$?
︡d4ba3302-3b63-4adf-83d8-f3445e913e88︡{"done":true,"md":"Explain this code line by line:\n\n- define function of one input\n- Docstring\n- sin(m) is the sage symbolic sin function\n- `(sin(m) for m in xrange(1,n))` is like a list comprehension but it is \"lazy\" so it doesn't waste memory -- you can iterate over it though.\n- `sum(...)` iterates over its input summing it up\n- `float(...)` converts its input to a float\n\n\nThrow out a mathematical question: Is there a simple formula for $\\sum_{0< m\\leq n} \\sin(m)$ similar to the formula\n$\\sum_{0< m<n} m = \\frac{n(n-1)}{2}$?"}
︠b94c79ad-f0ab-4c81-b2d8-a25ee4c9b0d7s︠
sin_sum(10000)  # this might feel slow if you try it!
︡a693770f-555b-4905-92de-5c504e8cffec︡{"stdout":"1.939505410680705"}︡{"stdout":"\n"}︡{"done":true}︡
︠f0ca33c1-932f-4be5-b290-19a7eb37b8e4i︠
%md
- Is it slow?
- Or is the network (or SMC) just slow at sending the answer back?
- If it is slow, why?

︡a48b3ea8-e6fb-4c0c-8dd9-996b59997bad︡{"done":true,"md":"- Is it slow?\n- Or is the network (or SMC) just slow at sending the answer back?\n- If it is slow, why?"}
︠078d23a5-2285-4fd4-a4f6-26cc02187d84i︠
%md
There are many ways to **time execution of code:**

1. *Very naive:* use a stopwatch.  Start the code running and start a stopwatch on your phone or whatever. Look at the wall clock.
1. *Python:* use the `time` module, and prints
1. *Python:* use the `timeit` module, and better prints
1. In worksheets: `%time` and `%timeit`
1. In Sage: `cputime` and `walltime` functions.
1. Using `%prun` in a terminal (so create a terminal and type "sage").  This is an ipython feature.

And there are also many, many ways to **speed up code**...
︡63f55785-98c4-4b91-b9d9-6f157d321f67︡{"done":true,"md":"There are many ways to **time execution of code:**\n\n1. *Very naive:* use a stopwatch.  Start the code running and start a stopwatch on your phone or whatever. Look at the wall clock.\n1. *Python:* use the `time` module, and prints\n1. *Python:* use the `timeit` module, and better prints\n1. In worksheets: `%time` and `%timeit`\n1. In Sage: `cputime` and `walltime` functions.\n1. Using `%prun` in a terminal (so create a terminal and type \"sage\").  This is an ipython feature.\n\nAnd there are also many, many ways to **speed up code**..."}
︠75dd621f-fb22-472e-a4da-04aa0090c2d8s︠
# Let's time the function in all these ways.

# 1. seems like a couple of seconds...?

# 2. time module

import time
t = time.time()   # number of seconds since "the epoch" (exercise: when was that?)
sin_sum(10000)
print "elapsed time", time.time() - t

# nice thing -- you can put time.time()'s anywhere in your code -- sort of like print statements -- for debugging performance issues..
︡242676de-6888-4c67-a834-0d22ee6ba936︡{"stdout":"1.939505410680705"}︡{"stdout":"\n"}︡{"stdout":"elapsed time 23.8433499336\n"}︡{"done":true}︡
︠1c171ebe-aa45-4dc4-871a-1aa638e8c2d5s︠
print time.time()
︡0d4f28d4-0285-45a2-85da-bb1c493feb1e︡{"stdout":"1486592141.29\n"}︡{"done":true}︡
︠5aedf163-bdfe-4726-9a13-a342393bd929i︠
%md 
3: timeit module: see https://docs.python.org/2/library/timeit.html
︡85b1f361-bb58-4dc5-830b-92caf13d61e4︡{"done":true,"md":"3: timeit module: see https://docs.python.org/2/library/timeit.html"}
︠65fae70f-8b5f-4822-bb00-10839136f8c9s︠
import timeit
timeit.timeit('sin_sum(1000)', number=50)
︡dbfc4604-bfa8-423a-adfa-df8bf283e69b︡{"stderr":"Error in lines 2-2\nTraceback (most recent call last):\n  File \"/projects/sage/sage-7.5/local/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 982, in execute\n    exec compile(block+'\\n', '', 'single') in namespace, locals\n  File \"\", line 1, in <module>\n  File \"/projects/sage/sage-7.5/local/lib/python/timeit.py\", line 237, in timeit\n    return Timer(stmt, setup, timer).timeit(number)\n  File \"/projects/sage/sage-7.5/local/lib/python/timeit.py\", line 202, in timeit\n    timing = self.inner(it, self.timer)\n  File \"<timeit-src>\", line 6, in inner\nNameError: global name 'sin_sum' is not defined\n"}︡{"done":true}︡
︠305211c0-d174-4226-9e2b-4ab1921bf6aei︠
%md
Hmm, let's check the documentation. Aha!
︡f5129d42-987f-4f89-ab27-9cc65fb3ec2f︡{"done":true,"md":"Hmm, let's check the documentation. Aha!"}
︠f75da578-b05e-416e-802c-0a72b768f54as︠
def tmp():
    return sin_sum(100)

timeit.timeit(tmp, number=500)
︡4755d47a-d8e3-4f9f-8dc4-76d6b92332d2︡{"stdout":"17.5108540058136"}︡{"stdout":"\n"}︡{"done":true}︡
︠5c58b32f-b2c3-4cec-8292-292d54649b18i︠
%md
4: %time, %timeit commands to SMC
︡9fc50db8-4162-49ed-92c5-763b9d0ed77b︡{"done":true,"md":"4: %time, %timeit commands to SMC"}
︠0034438a-a970-414b-918f-213d199bdd69s︠
%time
sin_sum(10000)
︡382c8c6b-a283-4621-8cf7-0cc07ca7e8db︡{"stderr":"Error in lines 1-1\nTraceback (most recent call last):\n  File \"/projects/sage/sage-7.5/local/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 982, in execute\n    exec compile(block+'\\n', '', 'single') in namespace, locals\n  File \"\", line 1, in <module>\n  File \"/projects/sage/sage-7.5/local/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 1027, in execute_with_code_decorators\n    code = code_decorator(code)\nTypeError: 'module' object is not callable\n"}︡{"done":true}︡
︠4e9aa281-e030-45b1-848e-4d9b2eef2794s︠
# the above doesn't work because we replace %time by doing "import time".... so do this:
reset('time')
︡a22f427b-d100-4099-af27-8fd60403c1eb︡{"done":true}︡
︠f49a85d3-0e14-4f17-a2e3-603647f326d5s︠
%time
sin_sum(10000)
︡52c536d2-ff91-4664-8d25-c5d795c70af8︡{"stdout":"1.939505410680705"}︡{"stdout":"\n"}︡{"stdout":"CPU time: 11.98 s, Wall time: 21.70 s\n"}︡{"done":true}︡{"stdout":"\n"}︡{"stdout":"CPU time: 11.98 s, Wall time: 21.70 s\n"}︡{"done":true}
︠f3e768e6-8e98-45d4-992a-96288a3f7746︠
%timeit
sin_sum(1000)
︡d9109a06-7468-4ae9-bf31-c04ab259f9ce︡{"stderr":"Error in lines 1-1\nTraceback (most recent call last):\n  File \"/projects/sage/sage-7.5/local/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 982, in execute\n    exec compile(block+'\\n', '', 'single') in namespace, locals\n  File \"\", line 1, in <module>\n  File \"/projects/sage/sage-7.5/local/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 1027, in execute_with_code_decorators\n    code = code_decorator(code)\nTypeError: 'module' object is not callable\n"}︡{"done":true}︡
︠d6f67a36-3045-43e5-bb07-2d6d58b60bbcs︠
reset('timeit')
%timeit
sin_sum(100)  # runs many times so give it smaller input!
︡66b5efff-084d-4d64-aaac-9450aa49bbdb︡{"stdout":"125 loops, best of 3: 4.96 ms per loop"}︡{"stdout":"\n"}︡{"done":true}︡
︠ee9d665b-4781-4040-b039-ab195dc392e8i︠
%md
5: `cputime` and `walltime` functions
︡1caa75ab-1c59-4f00-890b-45c8bd876e32︡{"done":true,"md":"5: `cputime` and `walltime` functions"}
︠e6328ec8-716f-4401-bc32-ab83e9fc3cd8s︠
t = cputime()
sin_sum(10000)
print cputime(t)
︡655bf87f-442c-478d-a657-af71bc9c5cea︡{"stdout":"1.939505410680705"}︡{"stdout":"\n"}︡{"stdout":"7.456\n"}︡{"done":true}︡
︠23770ea4-2f3a-4523-b4b9-5296849384a8s︠
t = walltime()
sin_sum(10000)
print walltime(t)
︡c0e59c35-86bd-431d-a3f7-8c132430a4d0︡{"stdout":"1.939505410680705"}︡{"stdout":"\n"}︡{"stdout":"16.6226541996\n"}︡{"done":true}︡
︠6000ddf5-9de9-4014-81dc-c2840340654di︠
%md
Discuss the difference between cpu and wall time.
︡ac57ffeb-a973-4926-baa7-76ecaaf57328︡{"done":true,"md":"Discuss the difference between cpu and wall time."}
︠d15fcf7b-8845-4690-8c21-26dab6a55001︠
%md
6: Using %prun in a terminal

- +New --> Terminal.
- type `sage` to start sage.
- Define the function.
- Use `%prun sin_sum(10000)`
︡82bd02e0-d4dd-444e-b6bf-546a121e30fd︡{"done":true,"md":"6: Using %prun in a terminal\n\n- +New --> Terminal.\n- type `sage` to start sage.\n- Define the function.\n- Use `%prun sin_sum(10000)`"}
︠8c7f4473-687b-4168-bd92-021fbc3963eai︠


%md
**Exercise:** Try to make sin_sum faster and time the result in several ways.  Ideas for how:

- replace `m` by `float(m)`
- replace `sin` by `math.sin`
︡9d3ac4cf-940b-42f5-a7ed-391ba90d4f1b︡{"done":true,"md":"**Exercise:** Try to make sin_sum faster and time the result in several ways.  Ideas for how:\n\n- replace `m` by `float(m)`\n- replace `sin` by `math.sin`"}
︠60c049df-d068-4cbc-bb0e-834193f018cd︠

︡26a4d2d7-f4eb-40ae-8017-121acee4b2ea︡
︠25c5d545-81e4-4f1d-b459-4abef3331063︠
︡4c492fc2-72ab-435c-89e4-2a8f891a60e6︡{"done":true}︡
︠47aa66f4-4f14-46f6-866b-a59ca09e53dc︠
def sin_sum(n):
    """
    Return the sum  sin(1) + sin(2) + ... + sin(n-1)
    """
    return float(sum(sin(m) for m in xrange(1,n)))

def sin_sum(n):
    """
    Return the sum  sin(1) + sin(2) + ... + sin(n-1)
    """
    return float(sum(math.sin(m) for m in xrange(1,n)))
︡55fc856b-aed1-4735-9f9e-7a4770909435︡{"done":true}︡
︠7f6ff77c-a862-421a-a240-0db66da79dba︠

︡d451644b-3524-40bd-8be3-1aff24e73095︡
︠c57d8d0a-b337-41b6-9d22-08ce886e2884i︠

%md
Of course, you can also try to make the function faster using mathematical knowledge! Can you find an explicit formula for $\sum_{0< m< n} \sin(m)$?
︡82c768a8-1430-47b4-a72c-f5ccd3f7eb6e︡{"done":true,"md":"Of course, you can also try to make the function faster using mathematical knowledge! Can you find an explicit formula for $\\sum_{0< m< n} \\sin(m)$?"}
︠097cc9c2-ea87-472c-b1fd-28bc2392a9ac︠









